<?php declare(strict_types=1);

namespace parkdown;

class Lexer {
	private array   $source;
	private ?string $fileName;
	
	public function __construct(string $sourceCode, ?string $fileName = null) {
		$this->fileName = $fileName;
		$unifiedSource  = str_replace(["\r\n", "\r"], "\n", $sourceCode);
		$unifiedSource  = str_replace("    ", "\t", $unifiedSource);
		$this->source   = explode("\n", trim($unifiedSource, "\n"));
	}

	public function tokenize() : array {
		$tokens = [];
		$row    = 1;
		$col    = 1;

		foreach ($this->source as $line) {
			if (strlen($line) < 1) {
				array_push($tokens, new Token(TokenType::EOL, "\n", [$row, 0, $this->fileName]));
				$row++;
				continue;
			}

			$buffer = "";
			$number = false;
			$col    = 1;

			$clearBuffer = function() use (&$buffer, &$tokens, $col, $row) {
				if (strlen($buffer) < 1)
					return;
				array_push($tokens, new Token(TokenType::TEXT, $buffer, [$col, $row, $this->fileName]));
				$buffer = "";
			};

			foreach(str_split($line) as $char) {
				if (is_numeric($char) && !$number) {
						$clearBuffer();
						$number = true;
				} else if (!is_numeric($char) && $number) {
					array_push($tokens, new Token(TokenType::NUMBER, $buffer, [$col, $row, $this->fileName]));
					$buffer = "";
					$number = false;
				}
				switch($char) {
					case '#':
						$clearBuffer();
						array_push($tokens, new Token(TokenType::HASH, $char, [$col, $row, $this->fileName]));
						break;
					case '*':
						$clearBuffer();
						array_push($tokens, new Token(TokenType::ASTERISK, $char, [$col, $row, $this->fileName]));
						break;
					case '.':
						$clearBuffer();
						array_push($tokens, new Token(TokenType::DOT, $char, [$col, $row, $this->fileName]));
						break;
					case '-':
						$clearBuffer();
						array_push($tokens, new Token(TokenType::MINUS, $char, [$col, $row, $this->fileName]));
						break;
					case '`':
						$clearBuffer();
						array_push($tokens, new Token(TokenType::BACKTICK, $char, [$col, $row, $this->fileName]));
						break;
					case '[':
						$clearBuffer();
						array_push($tokens, new Token(TokenType::LBRACKET, $char, [$col, $row, $this->fileName]));
						break;
					case ']':
						$clearBuffer();
						array_push($tokens, new Token(TokenType::RBRACKET, $char, [$col, $row, $this->fileName]));
						break;
					case '(':
						$clearBuffer();
						array_push($tokens, new Token(TokenType::LPAREN, $char, [$col, $row, $this->fileName]));
						break;
					case ')':
						$clearBuffer();
						array_push($tokens, new Token(TokenType::RPAREN, $char, [$col, $row, $this->fileName]));
						break;
					case '!':
						$clearBuffer();
						array_push($tokens, new Token(TokenType::BANG, $char, [$col, $row, $this->fileName]));
						break;
					case '|':
						$clearBuffer();
						array_push($tokens, new Token(TokenType::PIPE, $char, [$col, $row, $this->fileName]));
						break;
					case '\\':
						$clearBuffer();
						array_push($tokens, new Token(TokenType::BACKSLASH, $char, [$col, $row, $this->fileName]));
						break;
					case '>':
						$clearBuffer();
						array_push($tokens, new Token(TokenType::GT, $char, [$col, $row, $this->fileName]));
						break;
					case '	':
						$clearBuffer();
						array_push($tokens, new Token(TokenType::TAB, $char, [$col, $row, $this->fileName]));
						break;
					case ':':
						if (str_ends_with($buffer, "http") || str_ends_with($buffer, "https")) {
							$buffer .= $char;
							continue 2;
						}
						
						$clearBuffer();
						array_push($tokens, new Token(TokenType::COLON, $char, [$col, $row, $this->fileName]));
						break;
					case '{':
						$clearBuffer();
						array_push($tokens, new Token(TokenType::LBRACE, $char, [$col, $row, $this->fileName]));
						break;
					case '}':
						$clearBuffer();
						array_push($tokens, new Token(TokenType::RBRACE, $char, [$col, $row, $this->fileName]));
						break;
					case ' ':
						$clearBuffer();
						array_push($tokens, new Token(TokenType::SPACE, $char, [$col, $row, $this->fileName]));
						break;
					default:
						$buffer .= $char;
						break;
				}
			
				$col++;
			}
			$clearBuffer();
			array_push($tokens, new Token(TokenType::EOL, "\n", [$col, $row, $this->fileName]));
		
			$row++;
		}
		array_push($tokens, new Token(TokenType::EOF, "\0", [$col, $row, $this->fileName]));
		
		return $tokens;
	}
}